// Copyright (C) 1997, 1998 IAIK
// email: wplatzer@iaik.tu-graz.ac.at
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// $Header: /JigsawSSL/src/iaik/jigsaw/https/RequestServerCert.java 1     22.10.98 17:55 Asterbenz $

package iaik.jigsaw.https;

import java.io.*;
import java.security.*;
import java.util.*;

import iaik.asn1.*;
import iaik.asn1.structures.*;
import iaik.pkcs.pkcs8.EncryptedPrivateKeyInfo;
import iaik.pkcs.pkcs10.CertificateRequest;
import iaik.security.rsa.*;
import iaik.utils.KeyAndCertificate;
import iaik.utils.Base64OutputStream;
import iaik.x509.X509Certificate;

public class RequestServerCert {

  private final static boolean DEBUG = false;
  private final static boolean VCAFE = false;

  public static void generateRequest() throws Exception {
    String line;
    BufferedReader din = new BufferedReader(new InputStreamReader(System.in));

    System.out.println("Certificate Request Generation");
    System.out.println("==============================");
    System.out.println();
    System.out.println("Please enter the distinguished name to be used for this server.");
    System.out.println("If you leave a field blank it will not be included.");
    System.out.println();
    
    Name name = new Name();

    line = readNotEmpty(din, "Fully qualified DNS name of the host: ", "This is a required field, please enter the name!");
    name.addRDN(new RDN(ObjectID.commonName, line));

    System.out.print("Your Organization: ");
    line = din.readLine();
    if( ! line.equals("") ) {
      name.addRDN(new RDN(ObjectID.organization, line));
    }
    
    System.out.print("Your Organizational Unit: ");
    line = din.readLine();
    if( ! line.equals("") ) {
      name.addRDN(new RDN(ObjectID.organizationalUnit, line));
    }
    
    System.out.print("Your State or Province: ");
    line = din.readLine();
    if( ! line.equals("") ) {
      name.addRDN(new RDN(ObjectID.stateOrProvince, line));
    }
    
    System.out.print("Your Country: ");
    line = din.readLine();
    if( ! line.equals("") ) {
      name.addRDN(new RDN(ObjectID.country, line));
    }
    
    System.out.print("Your Locality: ");
    line = din.readLine();
    if( ! line.equals("") ) {
      name.addRDN(new RDN(ObjectID.locality, line));
    }
    
    System.out.println();
    System.out.println("The distinguished name you specified was:");
    System.out.println(name);
    System.out.println();
    
    line = choice(din, "Do you want to continue?", new String[] {"y", "n"});
    if( line.equals("n") ) {
      return;
    }
    
    System.out.println();
    System.out.println("Next the RSA keypair for the server will be generated.");
    line = choice(din, "Please specify the length of the key", new String[] {"512", "768", "1024", "2048"});
    int n = Integer.parseInt(line);
    
    System.out.println("A " + n + " bit keypair will now be generated, this may take a few minutes...");
    
    KeyPairGenerator kpg = KeyPairGenerator.getInstance("RSA", "IAIK");
    kpg.initialize(n);
    KeyPair keyPair = kpg.generateKeyPair();
    System.out.println("Keypair generation completed.");
    
    System.out.println();
    System.out.print("Finishing the certificate request...");
    CertificateRequest certRequest = new CertificateRequest(keyPair.getPublic(), name);
    certRequest.sign(AlgorithmID.md5WithRSAEncryption, keyPair.getPrivate());
    if( ! certRequest.verify() ) {
      System.out.println("Unexpected internal error generating certificate request, aborting!");
      return;
    }
    System.out.println("OK");
    System.out.println();
    
    PrivateKey privKey = keyPair.getPrivate();
    line = choice(din, "Do you want to encrypt the private key?", new String[] {"y", "n"});
    if( line.equals("y") ) {
      line = readNotEmpty(din, "Enter password for the private key: ", "Please enter a password!");
      EncryptedPrivateKeyInfo epki = new EncryptedPrivateKeyInfo(privKey);
      epki.encrypt(line, AlgorithmID.pbeWithSHAAnd3_KeyTripleDES_CBC, null);
      privKey = epki;
    }
    
    KeyAndCertificate kac = new KeyAndCertificate(privKey, new X509Certificate[0]);
    
    while( true ) {
      line = readNotEmpty(din, "Save filename for the PRIVATE key (KeyAndCertificate file): ", "Please enter a filename");
      try {
        kac.saveTo(line, ASN1.PEM);
        System.out.println("Private key saved ok.");
        System.out.println();
        break;
      } catch( IOException e ) {
        System.out.println("Error writing to " + line + ": " + e);
        System.out.println("Please enter a valid filename!");
      }
    }
    
    while( true ) {
      line = readNotEmpty(din, "Save filename for the certificate request: ", "Please enter a filename");
      try {
        OutputStream out = new FileOutputStream(line);
        out.write("-----BEGIN NEW CERTIFICATE REQUEST-----\n".getBytes());
        Base64OutputStream bout = new Base64OutputStream(out);
        certRequest.writeTo(bout);
        bout.flush();
        out.write("\n-----END NEW CERTIFICATE REQUEST-----\n".getBytes());
        out.flush();
        out.close();
        System.out.println("Certificate request saved ok.");
        System.out.println();
        break;
      } catch( IOException e ) {
        System.out.println("Error writing to " + line + ": " + e);
        System.out.println("Please enter a valid filename!");
      }
    }
    
    System.out.println("The certificate request has been generated successfully.");
    System.out.println("You can now send the contents of the file " + line);
    System.out.println("to a certification authority. When you have received the response with");
    System.out.println("the issued certificate execute step 2 of this program.");
  }
  
  public static void finishRequest() throws Exception {
    String line;
    BufferedReader din = new BufferedReader(new InputStreamReader(System.in));

    System.out.println("Certificate Installation");
    System.out.println("========================");
    System.out.println();
    
    KeyAndCertificate kac;
    PrivateKey privKey;
    X509Certificate[] certs;
    String kacFilename;
    String password = null;
    
    while( true ) {
      kacFilename = readNotEmpty(din, "Enter the filename of the private key generated in step 1: ", null);
      try {
        kac = new KeyAndCertificate(kacFilename);
        privKey = kac.getPrivateKey();
        System.out.println("OK");
        System.out.println();
        break;
      } catch( IOException e ) {
        System.out.println("Error reading private key: " + e);
        System.out.println("Please enter a valid filename!");
      }
    }
    
    if( privKey instanceof EncryptedPrivateKeyInfo ) {
      password = readNotEmpty(din, "Please enter the password for the private key: ", null);
      EncryptedPrivateKeyInfo epki = (EncryptedPrivateKeyInfo)privKey;
      try {
        epki.decrypt(password);
      } catch( Exception e ) {
        System.out.println("Unable to decrypt private key:");
        System.out.println(e);
        return;
      }
      privKey = epki.getPrivateKeyInfo();
    }
    
    while( true ) {
      line = readNotEmpty(din, "Enter the filename of the certificate file received from the CA: ", null);
      try {
        InputStream in = new FileInputStream(line);
        BufferedReader reader = new BufferedReader(new InputStreamReader(in));
        Vector certVector = new Vector();
        while( true ) {
          line = reader.readLine();
          if( line == null ) {
            break;
          }
          if( line.startsWith("-----BEGIN") ) {
            ByteArrayOutputStream bout = new ByteArrayOutputStream(8192);
            bout.write((line + "\n").getBytes());
            while( true ) {
              line = reader.readLine();
              if( line == null ) {
                throw new IOException("Unexpected end of file!");
              }
              bout.write((line + "\n").getBytes());
              if( line.startsWith("-----END") ) {
                break;
              }
            }
            X509Certificate cert = new X509Certificate(new ByteArrayInputStream(bout.toByteArray()));
            certVector.addElement(cert);
          }
        }
        int n = certVector.size();
        if( n == 0 ) {
          throw new IOException("No certificates found in the file!");
        }
        certs = new X509Certificate[n];
        certVector.copyInto(certs);
        System.out.println("OK, " + n + " certificates read");
        System.out.println();
        break;
      } catch( IOException e ) {
        System.out.println("Error reading certificate file: " + e);
      }
    }
    
    RSAPublicKey pubKey1 = (RSAPublicKey)((RSAPrivateKey)privKey).getPublicKey();
    RSAPublicKey pubKey2 = (RSAPublicKey)certs[0].getPublicKey();
//    System.out.println(pubKey1);
//    System.out.println(pubKey2);
    if( ! (pubKey1.getPublicExponent().equals(pubKey2.getPublicExponent()) && pubKey1.getModulus().equals(pubKey2.getModulus())) ) {
      System.out.println("The key private key and the public key in the certificate do not seem to match,");
      System.out.println("cannot continue.");
      return;
    }
    
    if( password != null ) {
      EncryptedPrivateKeyInfo epki = new EncryptedPrivateKeyInfo(privKey);
      epki.encrypt(password, AlgorithmID.pbeWithSHAAnd3_KeyTripleDES_CBC, null);
      privKey = epki;
    }
    
    kac = new KeyAndCertificate(privKey, certs);
    
    while( true ) {
      System.out.print("Please enter SAVE filename for the KeyAndCertificate file [" + kacFilename + "] ");
      line = din.readLine();
      if( line.equals("") ) {
        line = kacFilename;
      }
      try {
        kac.saveTo(line, ASN1.PEM);
        System.out.println("Private key and certificates saved successfully.");
        System.out.println();
        break;
      } catch( IOException e ) {
        System.out.println("Error saving key and certificate: " + e);
      }
    }
    
    System.out.println("You can now use the generated KeyAndCertificate file with your SSL server.");
  }
  
  public static void main0(String args[]) throws Exception {
    
    BufferedReader din = new BufferedReader(new InputStreamReader(System.in));
    
    System.out.println();
    System.out.println("Request a Server Certificate");
    System.out.println("============================");
    System.out.println();
    System.out.println("This program allows you to generate certificate requests");
    System.out.println("for iSaSiLk (Web) servers that can be sent to a Certification");
    System.out.println("Authority and to install the certificate received in response.");
    System.out.println("Currently only RSA keys are supported.");
    System.out.println();
    
    System.out.println("This procedure consists of two step:");
    System.out.println(" 1. Generate the certificate request");
    System.out.println(" 2. Install the certificate received from the CA");
    System.out.println();
    String choice = choice(din, "Select the step you want to execute", new String[] {"1", "2"});
    System.out.println();
    java.security.Security.addProvider(new iaik.security.provider.IAIK());
    if( choice.equals("1") ) {
      generateRequest();
    } else {
      finishRequest();
    }
  }

  public static String readNotEmpty(BufferedReader din, String text, String error) throws IOException {
    while( true ) {
      System.out.print(text);
      String line = din.readLine();
      if( !line.equals("") ) {
        return line;
      }
      if( error != null ) {
        System.out.println(error);
      }
    }
  }
  
  public static String choice(BufferedReader din, String text, String[] choices) throws IOException {
    while( true ) {
      System.out.print(text + " [");
      for( int i=0; i<choices.length; i++ ) {
        System.out.print(choices[i]);
        if( i < choices.length-1 ) {
          System.out.print("/");
        } else {
          System.out.print("] ");
        }
      }
      String line = din.readLine();
      for( int i=0; i<choices.length; i++ ) {
        if( line.equals(choices[i]) ) {
          return choices[i];
        }
      }
    }
  }
  
  public static void main(String args[]) {
    try {
      main0(args);
      exit(0);
    } catch( Throwable e ) {
      e.printStackTrace();
      exit(2);
    }
  }
  
  static void waitKey() {
    try {
      System.out.println("Hit the <RETURN> key.");
      do {
        System.in.read();
      } while( System.in.available() > 0 );
    } catch( java.io.IOException e ) {
      // ignore
    }
  }

  static void exit(int code) {
    if( VCAFE ) { // For Visual Cafe
      waitKey();
    }
    System.exit(code);
  }
}
